import React, {useMemo} from "react";
import {GlobalState, GlobalStateReducers, useGlobalStateReducers, useGlobalStateValue} from "../../GlobalState";

/**
 * Type of step key.
 */
export type StepKeyType = string;

/**
 * Steps global state data type.
 */
export interface StepsGlobalStateType
{
	/**
	 * Steps list, by their order of apparition.
	 */
	steps: { key: StepKeyType; title?: React.ReactNode; }[];

	/**
	 * The index of the current step in the steps list.
	 */
	currentStepIndex: number;
}

/**
 * The steps global state modification functions.
 */
class StepsGlobalStateReducers extends GlobalStateReducers<StepsGlobalStateType>
{
	/**
	 * Register a new step with its step key.
	 * @param newStepKey The step key to register.
	 */
	registerStep(newStepKey: StepKeyType, newStepTitle?: React.ReactNode): void
	{
		this.setState({
			// Add the given step key to the steps array, ensuring that it will be there only once.
			steps: [...this.state.steps.filter(({key: currentStepKey}) => currentStepKey != newStepKey), { key: newStepKey, title: newStepTitle }]
		});
	}

	/**
	 * Remove an old step with its step key.
	 * @param oldStepKey The step key to remove.
	 */
	removeStep(oldStepKey: StepKeyType): void
	{
		this.setState({
			// Remove the given step key from the steps array.
			steps: this.state.steps.filter(({key: currentStepKey}) => currentStepKey != oldStepKey)
		});
	}

	/**
	 * Set the current step from its key.
	 * @param stepKey The step key to set as the current one.
	 */
	setCurrentStep(stepKey: StepKeyType): void
	{
		// Get the new step index.
		const stepIndex = this.state.steps.findIndex(({key: currentStep}) => currentStep == stepKey);

		if (stepIndex >= 0)
		{ // If the new step index has been found, setting it as the current one.
			this.setState({
				currentStepIndex: stepIndex,
			});
		}
	}

	/**
	 * Get the next step key.
	 */
	getNextStep(): StepKeyType
	{
		return this.state.steps[this.state.currentStepIndex + 1]?.key ?? this.state.steps[0]?.key;
	}

	/**
	 * Get the previous step key.
	 */
	getPreviousStep(): StepKeyType
	{
		return this.state.steps[this.state.currentStepIndex - 1]?.key ?? this.state.steps[this.state.steps.length - 1]?.key;
	}
}

/**
 * The steps global state.
 */
export const stepsGlobalState = new GlobalState<StepsGlobalStateType, StepsGlobalStateReducers>({
	steps: [],
	currentStepIndex: 0,
}, new StepsGlobalStateReducers);

/**
 * Hook of the current step key in a steps component.
 */
export function useCurrentStepKey(): StepKeyType
{
	// Get the current steps global state value.
	const stepsState = useGlobalStateValue(stepsGlobalState);

	// Get the current step from the global state value.
	return stepsState.steps.length > 0
		? ( // If there are steps, getting the current one.
			stepsState.currentStepIndex < stepsState.steps.length
				? ( // If the index is correctly defined, trying to get the corresponding state key
					stepsState.currentStepIndex >= 0 ? (stepsState.steps[stepsState.currentStepIndex]?.key) : stepsState.steps[0]?.key
				)
				// The current index is too high, taking the last step as the current one.
				: stepsState.steps[stepsState.steps.length - 1]?.key
		)
		// There are no steps, returning none.
		: undefined;
}

/**
 * Steps navigator
 */
export interface StepsNavigator
{
	/**
	 * Go to the next step.
	 */
	next(): void;

	/**
	 * Return to the previous step.
	 */
	previous(): void;

	/**
	 * Go to the given step.
	 * @param stepKey The step key to join.
	 */
	set(stepKey: StepKeyType): void;
}

/**
 * Hook of the steps' navigator.
 */
export function useStepsNavigator(): StepsNavigator
{
	// Get the steps reducers instance.
	const stepsReducers = useGlobalStateReducers(stepsGlobalState);

	// Create the steps navigator object.
	return useMemo(() => ({
		next()
		{
			stepsReducers.setCurrentStep(stepsReducers.getNextStep());
		},

		previous()
		{
			stepsReducers.setCurrentStep(stepsReducers.getPreviousStep());
		},

		set(stepKey: StepKeyType)
		{
			stepsReducers.setCurrentStep(stepKey);
		},
	}), [stepsReducers]);
}
